
Analogs <- function(exp, obs, nanalogs = 3, fun_analog = "mean", lat_dim = "lat", lon_dim = "lon",
                    sdate_dim = "sdate", time_dim = "time", ncores = 1) {

  require(multiApply)
  require(CSTools)

  #-----------------------------------
  # Checkings
  #-----------------------------------
  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }

  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  if (!inherits(nanalogs, 'numeric')) {
    stop("Parameter 'nanalogs' must be of the class 'character'")
  }

  if (!inherits(fun_analog, 'character')) {
    stop("Parameter 'fun_analog' must be of the class 'character'")
  }

  # Do not allow synonims for lat (latitude), lon (longitude) and time (sdate) dimension names 
  if (is.na(match(lon_dim, names(dim(exp$data)))) | is.na(match(lon_dim, names(dim(obs$data))))) {
    stop("Missing longitude dimension in 'exp' and/or 'obs', or does not match the parameter
         'lon_dim'")
  } 

  if (is.na(match(lat_dim, names(dim(exp$data)))) | is.na(match(lat_dim, names(dim(obs$data))))) {
    stop("Missing latitude dimension in 'exp' and/or 'obs', or does not match the parameter
         'lat_dim'")
  }

  if (is.na(match(sdate_dim, names(dim(exp$data)))) | is.na(match(sdate_dim, names(dim(obs$data))))) {
    stop("Missing start date dimension in 'exp' and/or 'obs', or does not match the parameter
         'sdate_dim'")
  }

  if (is.na(match(time_dim, names(dim(exp$data)))) | is.na(match(time_dim, names(dim(obs$data))))) {
    stop("Missing time dimension in 'exp' and/or 'obs', or does not match the parameter
         'time_dim'")
  }

  # Ensure we have enough data to interpolate from high-res to coarse grid
  if ((attr(obs[[lat_dim]],"first_lat") > attr(exp[[lat_dim]],"first_lat")) | 
      (attr(obs[[lat_dim]],"last_lat") < attr(exp[[lat_dim]],"last_lat")) | 
      (attr(obs[[lon_dim]],"first_lon") > attr(exp[[lon_dim]],"first_lon")) |
      (attr(obs[[lon_dim]],"last_lon") < attr(exp[[lon_dim]],"last_lon"))) {
    stop("There are not enough data in 'obs'. Please to add more latitudes or ",
         "longitudes.")
  }

  # Select a function to apply to the analogs selected for a given observation
  stopifnot(fun_analog %in% c("mean", "wmean", "max", "min", "median"))

  #-----------------------------------
  # Interpolate high-res observations to the coarse grid
  #-----------------------------------
  grid_exp <- exp$source_files[1]  
  
  lonmin <- attr(exp[[lon_dim]],"first_lon")
  lonmax <- attr(exp[[lon_dim]],"last_lon")
  latmin <- attr(exp[[lat_dim]],"first_lat")
  latmax <- attr(exp[[lat_dim]],"last_lat")
  obs_interpolated <- Interpolation(exp = obs, target_grid = grid_exp, method_remap = "conservative",
                                    remap_region = c(lonmin, lonmax, latmin, latmax))
  lats <- obs[[lat_dim]]
  lons <- obs[[lon_dim]]

  #-----------------------------------
  # Reshape train and test
  #-----------------------------------
  res <- Apply(list(obs_interpolated$data, exp$data, obs$data), 
               target_dims = c(sdate_dim, time_dim, lat_dim, lon_dim), 
               fun = function(tr, te, ob) .analogs(train = tr, test = te, obs_hres = ob, k = nanalogs, 
               fun_analog = fun_analog), ncores = ncores)$output1

  #test <- exp$data[1,1,,1,,,]
  #train <- obs_interpolated$data[1,1,,,,]
  #obs_hres <- obs$data[1,1,,,,]
  #test <- aperm(test,c(2,1,3,4))
  #train <- aperm(train,c(3,1,2,4))
  #obs_hres <- aperm(obs_hres,c(3,1,2,4))  

  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = lons, lat = lats))
  
  return(res_s2dv)
}


# For each element in test, find the indices of the k nearest neigbhors in train
.analogs <- function(train, test, obs_hres, k, fun_analog) {
  # train, test, and obs_hres  dim: 4 dimensions sdate, time, lat and lon (in this order)
  require(FNN)
  require(multiApply)

  # Number of lats/lons of the high-resolution data
  space_dims_hres <- dim(obs_hres)[c(3,4)]
  nsdates <-  dim(train)[1]

  # Reformat train and test as an array with (time, points)
  train <- apply(train, c(1,2), as.vector); names(dim(train))[1] <- "space"  
  test <- apply(test, c(1,2), as.vector); names(dim(test))[1] <- "space"
  obs_hres <- apply(obs_hres, c(1,2), as.vector); names(dim(obs_hres))[1] <- "space"

  # Here we do cross-validation: the start date considered is removed from the training
  analogs <- lapply(1:nsdates, function(j) {
    tr <- apply(train[ , -j, ], 1, as.vector); names(dim(tr))[1] <- "time"
    te <- aperm(test[ , j, ])
    ob <- apply(obs_hres[ , -j, ], 1, as.vector); names(dim(ob))[1] <- "time"

    knn.ind <- get.knnx(tr, te, k)
   
    dist <- knn.ind$nn.dist
    idx <- knn.ind$nn.index
    names(dim(idx)) <- c("time", "analog"); names(dim(dist)) <- c("time", "analog")

    analogs <- Apply(idx, margins = "analog", fun = function(an) { ob[ an, ] })$output1
    dim(analogs) <- c(dim(analogs)[1], space_dims_hres, analog = k) 

    if (!is.null(fun_analog)) {
 
      margins_analogs <- names(dim(analogs))[names(dim(analogs)) != "analog"]

      if (fun_analog == "wmean") {
        weight <- 1 / dist
        analogs <- Apply(list(analogs, weight), margins = list(margins_analogs, "time"),
                         fun = function(x, w) weighted.mean(x, w))$output1
      } else {
        analogs <- Apply(analogs, margins = margins_analogs, fun = fun_analog)$output1
      }
    }
  })

  return(analogs)
}


